@@ -13,8 +13,8 @@
  *
  * You should have received a copy of the GNU Library General Public
  * License along with this library; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA 02111-1307, USA.
+ * Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
+ * Boston, MA 02110-1301, USA.
  */
 
 /* FIXME:
@@ -60,6 +60,7 @@ asf_packet_read_varlen_int (guint lentype_flags, guint lentype_bit_offset,
       val = GST_READ_UINT32_LE (*p_data);
       break;
     default:
+      val = 0;
       g_assert_not_reached ();
   }
 
@@ -132,19 +133,16 @@ gst_asf_payload_queue_for_stream (GstASFDemux * demux, AsfPayload * payload,
 {
   GST_DEBUG_OBJECT (demux, "Got payload for stream %d ts:%" GST_TIME_FORMAT,
       stream->id, GST_TIME_ARGS (payload->ts));
-  /* remember the first timestamp in the stream */
-  if (G_UNLIKELY (!GST_CLOCK_TIME_IS_VALID (demux->first_ts) &&
-          GST_CLOCK_TIME_IS_VALID (payload->ts))) {
-    GST_DEBUG_OBJECT (demux, "first ts: %" GST_TIME_FORMAT,
-        GST_TIME_ARGS (payload->ts));
-    demux->first_ts = payload->ts;
-  }
 
-  /* make timestamps start from 0 */
-  if (G_LIKELY (demux->first_ts < payload->ts))
-    payload->ts -= demux->first_ts;
-  else
-    payload->ts = 0;
+  /* make timestamps start from 0; first_ts will be determined during activation (once we have enough data),
+     which will also update ts of all packets queued before we knew first_ts;  */
+  if (G_LIKELY (GST_CLOCK_TIME_IS_VALID (demux->first_ts)
+          && GST_CLOCK_TIME_IS_VALID (payload->ts))) {
+    if (payload->ts > demux->first_ts)
+      payload->ts -= demux->first_ts;
+    else
+      payload->ts = 0;
+  }
 
   /* remove any incomplete payloads that will never be completed */
   while (stream->payloads->len > 0) {
@@ -194,10 +192,10 @@ gst_asf_payload_queue_for_stream (GstASFDemux * demux, AsfPayload * payload,
 
   /* remember the first queued timestamp for the segment */
   if (G_UNLIKELY (!GST_CLOCK_TIME_IS_VALID (demux->segment_ts) &&
-          GST_CLOCK_TIME_IS_VALID (payload->ts))) {
+          GST_CLOCK_TIME_IS_VALID (demux->first_ts))) {
     GST_DEBUG_OBJECT (demux, "segment ts: %" GST_TIME_FORMAT,
-        GST_TIME_ARGS (payload->ts));
-    demux->segment_ts = payload->ts;
+        GST_TIME_ARGS (demux->first_ts));
+    demux->segment_ts = demux->first_ts;
     /* always note, but only determines segment when streaming */
     if (demux->streaming)
       gst_segment_do_seek (&demux->segment, demux->in_segment.rate,
@@ -214,29 +212,35 @@ asf_payload_parse_replicated_data_extensions (AsfStream * stream,
 {
   AsfPayloadExtension *ext;
   guint off;
+  guint16 ext_len;
 
   if (!stream->ext_props.valid || stream->ext_props.payload_extensions == NULL)
     return;
 
   off = 8;
   for (ext = stream->ext_props.payload_extensions; ext->len > 0; ++ext) {
-    if (G_UNLIKELY (off + ext->len > payload->rep_data_len)) {
+    ext_len = ext->len;
+    if (ext_len == 0xFFFF) {    /* extension length is determined by first two bytes in replicated data */
+      ext_len = GST_READ_UINT16_LE (payload->rep_data + off);
+      off += 2;
+    }
+    if (G_UNLIKELY (off + ext_len > payload->rep_data_len)) {
       GST_WARNING ("not enough replicated data for defined extensions");
       return;
     }
     switch (ext->id) {
       case ASF_PAYLOAD_EXTENSION_DURATION:
-        if (G_LIKELY (ext->len == 2)) {
+        if (G_LIKELY (ext_len == 2)) {
           guint16 tdur = GST_READ_UINT16_LE (payload->rep_data + off);
           /* packet durations of 1ms are mostly invalid */
           if (tdur != 1)
             payload->duration = tdur * GST_MSECOND;
         } else {
-          GST_WARNING ("unexpected DURATION extensions len %u", ext->len);
+          GST_WARNING ("unexpected DURATION extensions len %u", ext_len);
         }
         break;
       case ASF_PAYLOAD_EXTENSION_SYSTEM_CONTENT:
-        if (G_LIKELY (ext->len == 1)) {
+        if (G_LIKELY (ext_len == 1)) {
           guint8 data = payload->rep_data[off];
 
           payload->interlaced = data & 0x1;
@@ -245,24 +249,34 @@ asf_payload_parse_replicated_data_extensions (AsfStream * stream,
           GST_DEBUG ("SYSTEM_CONTENT: interlaced:%d, rff:%d, tff:%d",
               payload->interlaced, payload->rff, payload->tff);
         } else {
-          GST_WARNING ("unexpected SYSTEM_CONTE extensions len %u", ext->len);
+          GST_WARNING ("unexpected SYSTEM_CONTE extensions len %u", ext_len);
         }
         break;
       case ASF_PAYLOAD_EXTENSION_SYSTEM_PIXEL_ASPECT_RATIO:
-        if (G_LIKELY (ext->len == 2)) {
+        if (G_LIKELY (ext_len == 2)) {
           payload->par_x = payload->rep_data[off];
           payload->par_y = payload->rep_data[off + 1];
           GST_DEBUG ("PAR %d / %d", payload->par_x, payload->par_y);
         } else {
           GST_WARNING ("unexpected SYSTEM_PIXEL_ASPECT_RATIO extensions len %u",
-              ext->len);
+              ext_len);
         }
         break;
+      case ASF_PAYLOAD_EXTENSION_TIMING:
+      {
+        /* dvr-ms timing - this will override packet timestamp */
+        guint64 time = GST_READ_UINT64_LE (payload->rep_data + off + 8);
+        if (time != 0xFFFFFFFFFFFFFFFF)
+          payload->ts = time * 100;
+        else
+          payload->ts = GST_CLOCK_TIME_NONE;
+      }
+        break;
       default:
-        GST_WARNING ("UNKNOWN PAYLOAD EXTENSION !");
+        GST_LOG ("UNKNOWN PAYLOAD EXTENSION!");
         break;
     }
-    off += ext->len;
+    off += ext_len;
   }
 }
 
@@ -342,8 +356,10 @@ gst_asf_demux_parse_payload (GstASFDemux * demux, AsfPacket * packet,
   stream = gst_asf_demux_get_stream (demux, stream_num);
 
   if (G_UNLIKELY (stream == NULL)) {
-    GST_WARNING_OBJECT (demux, "Payload for unknown stream %u, skipping",
-        stream_num);
+    if (gst_asf_demux_is_unknown_stream (demux, stream_num)) {
+      GST_WARNING_OBJECT (demux, "Payload for unknown stream %u, skipping",
+          stream_num);
+    }
     if (*p_size < payload_len) {
       *p_data += *p_size;
       *p_size = 0;
@@ -493,14 +509,14 @@ gst_asf_demux_parse_payload (GstASFDemux * demux, AsfPacket * packet,
   return TRUE;
 }
 
-gboolean
+GstAsfDemuxParsePacketError
 gst_asf_demux_parse_packet (GstASFDemux * demux, GstBuffer * buf)
 {
   AsfPacket packet = { 0, };
   GstMapInfo map;
   const guint8 *data;
   gboolean has_multiple_payloads;
-  gboolean ret = TRUE;
+  GstAsfDemuxParsePacketError ret = GST_ASF_DEMUX_PARSE_PACKET_ERROR_NONE;
   guint8 ec_flags, flags1;
   guint size;
 
@@ -510,8 +526,11 @@ gst_asf_demux_parse_packet (GstASFDemux * demux, GstBuffer * buf)
   GST_LOG_OBJECT (demux, "Buffer size: %u", size);
 
   /* need at least two payload flag bytes, send time, and duration */
-  if (G_UNLIKELY (size < 2 + 4 + 2))
-    goto short_packet;
+  if (G_UNLIKELY (size < 2 + 4 + 2)) {
+    GST_WARNING_OBJECT (demux, "Packet size is < 8");
+    ret = GST_ASF_DEMUX_PARSE_PACKET_ERROR_RECOVERABLE;
+    goto done;
+  }
 
   packet.buf = buf;
   /* evidently transient */
@@ -534,8 +553,11 @@ gst_asf_demux_parse_packet (GstASFDemux * demux, GstBuffer * buf)
     GST_LOG_OBJECT (demux, "packet has error correction (%u bytes)", ec_len);
 
     /* still need at least two payload flag bytes, send time, and duration */
-    if (size <= (1 + ec_len) + 2 + 4 + 2)
-      goto short_packet;
+    if (size <= (1 + ec_len) + 2 + 4 + 2) {
+      GST_WARNING_OBJECT (demux, "Packet size is < 8 with Error Correction");
+      ret = GST_ASF_DEMUX_PARSE_PACKET_ERROR_FATAL;
+      goto done;
+    }
 
     data += 1 + ec_len;
     size -= 1 + ec_len;
@@ -556,8 +578,11 @@ gst_asf_demux_parse_packet (GstASFDemux * demux, GstBuffer * buf)
 
   packet.padding = asf_packet_read_varlen_int (flags1, 3, &data, &size);
 
-  if (G_UNLIKELY (size < 6))
-    goto short_packet;
+  if (G_UNLIKELY (size < 6)) {
+    GST_WARNING_OBJECT (demux, "Packet size is < 6");
+    ret = GST_ASF_DEMUX_PARSE_PACKET_ERROR_FATAL;
+    goto done;
+  }
 
   packet.send_time = GST_READ_UINT32_LE (data) * GST_MSECOND;
   packet.duration = GST_READ_UINT16_LE (data + 4) * GST_MSECOND;
@@ -575,8 +600,11 @@ gst_asf_demux_parse_packet (GstASFDemux * demux, GstBuffer * buf)
   GST_LOG_OBJECT (demux, "duration         : %" GST_TIME_FORMAT,
       GST_TIME_ARGS (packet.duration));
 
-  if (G_UNLIKELY (packet.padding == (guint) - 1 || size < packet.padding))
-    goto short_packet;
+  if (G_UNLIKELY (packet.padding == (guint) - 1 || size < packet.padding)) {
+    GST_WARNING_OBJECT (demux, "No padding, or padding bigger than buffer");
+    ret = GST_ASF_DEMUX_PARSE_PACKET_ERROR_RECOVERABLE;
+    goto done;
+  }
 
   size -= packet.padding;
 
@@ -588,7 +616,9 @@ gst_asf_demux_parse_packet (GstASFDemux * demux, GstBuffer * buf)
         "adjusting available data size");
     if (size < demux->packet_size - packet.length) {
       /* the buffer is smaller than the implicit padding */
-      goto short_packet;
+      GST_WARNING_OBJECT (demux, "Buffer is smaller than the implicit padding");
+      ret = GST_ASF_DEMUX_PARSE_PACKET_ERROR_RECOVERABLE;
+      goto done;
     } else {
       /* subtract the implicit padding */
       size -= (demux->packet_size - packet.length);
@@ -598,8 +628,11 @@ gst_asf_demux_parse_packet (GstASFDemux * demux, GstBuffer * buf)
   if (has_multiple_payloads) {
     guint i, num, lentype;
 
-    if (G_UNLIKELY (size < 1))
-      goto short_packet;
+    if (G_UNLIKELY (size < 1)) {
+      GST_WARNING_OBJECT (demux, "No room more in buffer");
+      ret = GST_ASF_DEMUX_PARSE_PACKET_ERROR_RECOVERABLE;
+      goto done;
+    }
 
     num = (GST_READ_UINT8 (data) & 0x3F) >> 0;
     lentype = (GST_READ_UINT8 (data) & 0xC0) >> 6;
@@ -613,26 +646,23 @@ gst_asf_demux_parse_packet (GstASFDemux * demux, GstBuffer * buf)
       GST_LOG_OBJECT (demux, "Parsing payload %u/%u, size left: %u", i + 1, num,
           size);
 
-      ret = gst_asf_demux_parse_payload (demux, &packet, lentype, &data, &size);
-
-      if (G_UNLIKELY (!ret)) {
+      if (G_UNLIKELY (!gst_asf_demux_parse_payload (demux, &packet, lentype,
+                  &data, &size))) {
         GST_WARNING_OBJECT (demux, "Failed to parse payload %u/%u", i + 1, num);
+        ret = GST_ASF_DEMUX_PARSE_PACKET_ERROR_FATAL;
         break;
       }
     }
   } else {
     GST_LOG_OBJECT (demux, "Parsing single payload");
-    ret = gst_asf_demux_parse_payload (demux, &packet, -1, &data, &size);
+    if (G_UNLIKELY (!gst_asf_demux_parse_payload (demux, &packet, -1, &data,
+                &size))) {
+      GST_WARNING_OBJECT (demux, "Failed to parse payload");
+      ret = GST_ASF_DEMUX_PARSE_PACKET_ERROR_RECOVERABLE;
+    }
   }
 
+done:
   gst_buffer_unmap (buf, &map);
   return ret;
-
-/* ERRORS */
-short_packet:
-  {
-    gst_buffer_unmap (buf, &map);
-    GST_WARNING_OBJECT (demux, "Short packet!");
-    return FALSE;
-  }
 }
